# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Declarative tests framework.

This module provides fixtures and utilities for testing Airbyte sources and destinations
in a declarative way.
"""



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for connector test suites."""

from __future__ import annotations

import abc
import inspect
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal, cast

import pytest
import yaml
from boltons.typeutils import classproperty

from airbyte_cdk import Connector
from airbyte_cdk.models import (
    AirbyteMessage,
    Type,
)
from airbyte_cdk.test import entrypoint_wrapper
from airbyte_cdk.test.declarative.models import (
    ConnectorTestScenario,
)
from airbyte_cdk.test.declarative.utils.job_runner import IConnector, run_test_job

ACCEPTANCE_TEST_CONFIG = "acceptance-test-config.yml"
MANIFEST_YAML = "manifest.yaml"


class JavaClass(str):
    """A string that represents a Java class."""


class DockerImage(str):
    """A string that represents a Docker image."""


class RunnableConnector(abc.ABC):
    """A connector that can be run in a test scenario."""

    @abc.abstractmethod
    def launch(cls, args: list[str] | None) -> None: ...


def generate_tests(metafunc: pytest.Metafunc) -> None:
    """
    A helper for pytest_generate_tests hook.

    If a test method (in a class subclassed from our base class)
    declares an argument 'instance', this function retrieves the
    'scenarios' attribute from the test class and parametrizes that
    test with the values from 'scenarios'.

    ## Usage

    ```python
    from airbyte_cdk.test.declarative.test_suites.connector_base import (
        generate_tests,
        ConnectorTestSuiteBase,
    )

    def pytest_generate_tests(metafunc):
        generate_tests(metafunc)

    class TestMyConnector(ConnectorTestSuiteBase):
        ...

    ```
    """
    # Check if the test function requires an 'instance' argument
    if "instance" in metafunc.fixturenames:
        # Retrieve the test class
        test_class = metafunc.cls
        if test_class is None:
            raise ValueError("Expected a class here.")
        # Get the 'scenarios' attribute from the class
        scenarios_attr = getattr(test_class, "get_scenarios", None)
        if scenarios_attr is None:
            raise ValueError(
                f"Test class {test_class} does not have a 'scenarios' attribute. "
                "Please define the 'scenarios' attribute in the test class."
            )

        scenarios = test_class.get_scenarios()
        ids = [str(scenario) for scenario in scenarios]
        metafunc.parametrize("instance", scenarios, ids=ids)


class ConnectorTestSuiteBase(abc.ABC):
    """Base class for connector test suites."""

    connector: type[IConnector] | Callable[[], IConnector] | None = None
    """The connector class or a factory function that returns an instance of IConnector."""

    @classmethod
    def get_test_class_dir(cls) -> Path:
        """Get the file path that contains the class."""
        module = sys.modules[cls.__module__]
        # Get the directory containing the test file
        return Path(inspect.getfile(module)).parent

    @classmethod
    def create_connector(
        cls,
        scenario: ConnectorTestScenario,
    ) -> IConnector:
        """Instantiate the connector class."""
        connector = cls.connector  # type: ignore
        if connector:
            if callable(connector) or isinstance(connector, type):
                # If the connector is a class or factory function, instantiate it:
                return cast(IConnector, connector())  # type: ignore [redundant-cast]

        # Otherwise, we can't instantiate the connector. Fail with a clear error message.
        raise NotImplementedError(
            "No connector class or connector factory function provided. "
            "Please provide a class or factory function in `cls.connector`, or "
            "override `cls.create_connector()` to define a custom initialization process."
        )

    def run_test_scenario(
        self,
        verb: Literal["read", "check", "discover"],
        test_scenario: ConnectorTestScenario,
        *,
        catalog: dict[str, Any] | None = None,
    ) -> entrypoint_wrapper.EntrypointOutput:
        """Run a test job from provided CLI args and return the result."""
        return run_test_job(
            self.create_connector(test_scenario),
            verb,
            test_instance=test_scenario,
            catalog=catalog,
        )

    # Test Definitions

    def test_check(
        self,
        instance: ConnectorTestScenario,
    ) -> None:
        """Run `connection` acceptance tests."""
        result = self.run_test_scenario(
            "check",
            test_scenario=instance,
        )
        conn_status_messages: list[AirbyteMessage] = [
            msg for msg in result._messages if msg.type == Type.CONNECTION_STATUS
        ]  # noqa: SLF001  # Non-public API
        assert len(conn_status_messages) == 1, (
            f"Expected exactly one CONNECTION_STATUS message. Got: {result._messages}"
        )

    @classmethod
    def get_connector_root_dir(cls) -> Path:
        """Get the root directory of the connector."""
        for parent in cls.get_test_class_dir().parents:
            if (parent / MANIFEST_YAML).exists():
                return parent
            if (parent / ACCEPTANCE_TEST_CONFIG).exists():
                return parent
            if parent.name == "airbyte_cdk":
                break
        # If we reach here, we didn't find the manifest file in any parent directory
        # Check if the manifest file exists in the current directory
        for parent in Path.cwd().parents:
            if (parent / MANIFEST_YAML).exists():
                return parent
            if (parent / ACCEPTANCE_TEST_CONFIG).exists():
                return parent
            if parent.name == "airbyte_cdk":
                break

        raise FileNotFoundError(
            "Could not find connector root directory relative to "
            f"'{str(cls.get_test_class_dir())}' or '{str(Path.cwd())}'."
        )

    @classproperty
    def acceptance_test_config_path(cls) -> Path:
        """Get the path to the acceptance test config file."""
        result = cls.get_connector_root_dir() / ACCEPTANCE_TEST_CONFIG
        if result.exists():
            return result

        raise FileNotFoundError(f"Acceptance test config file not found at: {str(result)}")

    @classmethod
    def get_scenarios(
        cls,
    ) -> list[ConnectorTestScenario]:
        """Get acceptance tests for a given category.

        This has to be a separate function because pytest does not allow
        parametrization of fixtures with arguments from the test class itself.
        """
        category = "connection"
        all_tests_config = yaml.safe_load(cls.acceptance_test_config_path.read_text())
        if "acceptance_tests" not in all_tests_config:
            raise ValueError(
                f"Acceptance tests config not found in {cls.acceptance_test_config_path}."
                f" Found only: {str(all_tests_config)}."
            )
        if category not in all_tests_config["acceptance_tests"]:
            return []
        if "tests" not in all_tests_config["acceptance_tests"][category]:
            raise ValueError(f"No tests found for category {category}")

        tests_scenarios = [
            ConnectorTestScenario.model_validate(test)
            for test in all_tests_config["acceptance_tests"][category]["tests"]
            if "iam_role" not in test["config_path"]
        ]
        connector_root = cls.get_connector_root_dir().absolute()
        for test in tests_scenarios:
            if test.config_path:
                test.config_path = connector_root / test.config_path
            if test.configured_catalog_path:
                test.configured_catalog_path = connector_root / test.configured_catalog_path

        return tests_scenarios



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Declarative test suites.

Here we have base classes for a robust set of declarative connector test suites.
"""

from airbyte_cdk.test.declarative.test_suites.connector_base import (
    ConnectorTestScenario,
    ConnectorTestSuiteBase,
    generate_tests,
)
from airbyte_cdk.test.declarative.test_suites.declarative_sources import (
    DeclarativeSourceTestSuite,
)
from airbyte_cdk.test.declarative.test_suites.destination_base import DestinationTestSuiteBase
from airbyte_cdk.test.declarative.test_suites.source_base import SourceTestSuiteBase

__all__ = [
    "ConnectorTestScenario",
    "ConnectorTestSuiteBase",
    "DeclarativeSourceTestSuite",
    "DestinationTestSuiteBase",
    "SourceTestSuiteBase",
    "generate_tests",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for destination test suites."""

from airbyte_cdk.test.declarative.test_suites.connector_base import ConnectorTestSuiteBase


class DestinationTestSuiteBase(ConnectorTestSuiteBase):
    """Base class for destination test suites.

    This class provides a base set of functionality for testing destination connectors, and it
    inherits all generic connector tests from the `ConnectorTestSuiteBase` class.
    """



"""Auto-use fixtures for pytest.

WARNING: Importing this module will automatically apply these fixtures. If you want to selectively
enable fixtures in a different context, you can import directly from the `fixtures.general` module.


Usage:

```python
from airbyte_cdk.test.fixtures import auto
# OR
from airbyte_cdk.test.fixtures.auto import *
```
"""



import tempfile
import uuid
from dataclasses import asdict
from pathlib import Path
from typing import Any, Callable, Literal

import orjson
from typing_extensions import Protocol, runtime_checkable

from airbyte_cdk.models import (
    ConfiguredAirbyteCatalog,
    Status,
)
from airbyte_cdk.test import entrypoint_wrapper
from airbyte_cdk.test.declarative.models import (
    ConnectorTestScenario,
)


@runtime_checkable
class IConnector(Protocol):
    """A connector that can be run in a test scenario."""

    def launch(self, args: list[str] | None) -> None:
        """Launch the connector with the given arguments."""
        ...


def run_test_job(
    connector: IConnector | type[IConnector] | Callable[[], IConnector],
    verb: Literal["read", "check", "discover"],
    test_instance: ConnectorTestScenario,
    *,
    catalog: ConfiguredAirbyteCatalog | dict[str, Any] | None = None,
) -> entrypoint_wrapper.EntrypointOutput:
    """Run a test job from provided CLI args and return the result."""
    if not connector:
        raise ValueError("Connector is required")

    if catalog and isinstance(catalog, ConfiguredAirbyteCatalog):
        # Convert the catalog to a dict if it's already a ConfiguredAirbyteCatalog.
        catalog = asdict(catalog)

    connector_obj: IConnector
    if isinstance(connector, type) or callable(connector):
        # If the connector is a class or a factory lambda, instantiate it.
        connector_obj = connector()
    elif (
        isinstance(
            connector,
            IConnector,
        )
        or True
    ):  # TODO: Get a valid protocol check here
        connector_obj = connector
    else:
        raise ValueError(
            f"Invalid connector input: {type(connector)}",
        )

    args: list[str] = [verb]
    if test_instance.config_path:
        args += ["--config", str(test_instance.config_path)]
    elif test_instance.config_dict:
        config_path = (
            Path(tempfile.gettempdir()) / "airbyte-test" / f"temp_config_{uuid.uuid4().hex}.json"
        )
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(orjson.dumps(test_instance.config_dict).decode())
        args += ["--config", str(config_path)]

    catalog_path: Path | None = None
    if verb not in ["discover", "check"]:
        # We need a catalog for read.
        if catalog:
            # Write the catalog to a temp json file and pass the path to the file as an argument.
            catalog_path = (
                Path(tempfile.gettempdir())
                / "airbyte-test"
                / f"temp_catalog_{uuid.uuid4().hex}.json"
            )
            catalog_path.parent.mkdir(parents=True, exist_ok=True)
            catalog_path.write_text(orjson.dumps(catalog).decode())
        elif test_instance.configured_catalog_path:
            catalog_path = Path(test_instance.configured_catalog_path)

        if catalog_path:
            args += ["--catalog", str(catalog_path)]

    # This is a bit of a hack because the source needs the catalog early.
    # Because it *also* can fail, we have ot redundantly wrap it in a try/except block.

    result: entrypoint_wrapper.EntrypointOutput = entrypoint_wrapper._run_command(  # noqa: SLF001  # Non-public API
        source=connector_obj,  # type: ignore [arg-type]
        args=args,
        expecting_exception=test_instance.expect_exception,
    )
    if result.errors and not test_instance.expect_exception:
        raise AssertionError(
            "\n\n".join(
                [str(err.trace.error).replace("\\n", "\n") for err in result.errors if err.trace],
            )
        )

    if verb == "check":
        # Check is expected to fail gracefully without an exception.
        # Instead, we assert that we have a CONNECTION_STATUS message with
        # a failure status.
        assert not result.errors, "Expected no errors from check. Got:\n" + "\n".join(
            [str(error) for error in result.errors]
        )
        assert len(result.connection_status_messages) == 1, (
            "Expected exactly one CONNECTION_STATUS message. Got "
            f"{len(result.connection_status_messages)}:\n"
            + "\n".join([str(msg) for msg in result.connection_status_messages])
        )
        if test_instance.expect_exception:
            conn_status = result.connection_status_messages[0].connectionStatus
            assert conn_status, (
                "Expected CONNECTION_STATUS message to be present. Got: \n"
                + "\n".join([str(msg) for msg in result.connection_status_messages])
            )
            assert conn_status.status == Status.FAILED, (
                "Expected CONNECTION_STATUS message to be FAILED. Got: \n"
                + "\n".join([str(msg) for msg in result.connection_status_messages])
            )

        return result

    # For all other verbs, we assert check that an exception is raised (or not).
    if test_instance.expect_exception:
        if not result.errors:
            raise AssertionError("Expected exception but got none.")

        return result
    if result.errors:
        raise AssertionError(
            "\n\n".join(
                [
                    str(err.trace.error).replace(
                        "\\n",
                        "\n",
                    )
                    for err in result.errors
                    if err.trace
                ],
            )
        )

    return result



"""Global pytest configuration for the Airbyte CDK tests."""

from pathlib import Path
from typing import cast

import pytest


def pytest_collect_file(parent: pytest.Module | None, path: Path) -> pytest.Module | None:
    """Collect test files based on their names."""
    if path.name == "test_connector.py":
        return cast(pytest.Module, pytest.Module.from_parent(parent, path=path))

    return None


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line("markers", "connector: mark test as a connector test")


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--run-connector",
        action="store_true",
        default=False,
        help="run connector tests",
    )


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]) -> None:
    if config.getoption("--run-connector"):
        return
    skip_connector = pytest.mark.skip(reason="need --run-connector option to run")
    for item in items:
        if "connector" in item.keywords:
            item.add_marker(skip_connector)


def pytest_runtest_setup(item: pytest.Item) -> None:
    # This hook is called before each test function is executed
    print(f"Setting up test: {item.name}")


def pytest_runtest_teardown(item: pytest.Item, nextitem: pytest.Item | None) -> None:
    # This hook is called after each test function is executed
    print(f"Tearing down test: {item.name}")
